"""Store various application values."""
import os


class AppKeys:
    """Store application keys in one place."""

    GITHUB_TOKEN = 'GITHUB_TOKEN'
    VAULT_ADDR = 'VAULT_ADDR'

    IMAGE_API_RELATIVE_URL = 'image_api_relative_url'
    IMAGE_ENHANCE_API_RELATIVE_URL = 'image_enhance_api_relative_url'


class SecretPaths:
    """Store secret paths used against the vault service."""

    TOKEN_ENDPOINT = os.environ['TOKEN_ENDPOINT']
    CLIENT_ID_API = os.environ['CLIENT_ID_API']
    CLIENT_SECRET_API = os.environ['CLIENT_SECRET_API']
    LOUVRE_DOMAIN = os.environ['LOUVRE_DOMAIN']

    # Against PROD
    TOKEN_ENDPOINT_PROD = os.environ['TOKEN_ENDPOINT_PROD']
    CLIENT_ID_API_PROD = os.environ['CLIENT_ID_API_PROD']
    CLIENT_SECRET_API_PROD = os.environ['CLIENT_SECRET_API_PROD']
    LOUVRE_DOMAIN_PROD = os.environ['LOUVRE_DOMAIN_PROD']

    ISSUER = os.environ['ISSUER']


class ImageVariants:
    """Allowed values for the image variant parameter."""

    STANDARD = 'standard'
    THUMBNAIL = 'thumbnail'
    ORIGINAL = 'original'
    DEFAULT = STANDARD



from elvia_louvre.data_models import UpdateMetadataRequest
from io import BytesIO
from PIL import Image
import requests

from .config import Config
from .enums import ImageVariants
from .image_api import ImageAPI
from .image_enhance_api import ImageEnhanceAPI


class LouvreClient:
    """Python client to interact with APIs in the Louvre system."""
    def __init__(self):
        Config.connect()

    def get_image_data(self,
                       image_id: str,
                       using_production_images: bool = False):
        """
        Return an object with image metadata retrieved from ImageAPI.
        
        Parameters
        ----------
        image_id : str
        using_production_images : bool
            Whether to search in production.

        Returns
        -------
        ImageData

        Raises
        ------      
        LouvreImageNotFound,
        LouvreKeyError,
        LouvreQueryError,
        RequestException,
        """
        return ImageAPI.get_image_data(
            img_id=image_id, using_production_images=using_production_images)

    def get_image_url(self,
                      image_id: str,
                      image_variant: str = ImageVariants.DEFAULT,
                      using_production_images: bool = False) -> str:
        """
        Return the URL of an image variant.

        Parameters
        ----------
        image_id : str
        image_variant : str
            Accepted values can be found in the ImageVariants class.
        using_production_images : bool
            Whether to search in production.        

        Returns
        -------
        str

        Raises
        ------
        LouvreImageNotFound,
        LouvreKeyError,
        LouvreQueryError,
        LouvreInvalidImageVariant,
        RequestException,
        """
        return ImageAPI.get_image_sasuri(
            img_id=image_id,
            image_variant=image_variant,
            using_production_images=using_production_images)

    def get_image(self,
                  image_id: str,
                  image_variant: str = ImageVariants.DEFAULT,
                  using_production_images: bool = False) -> Image:
        """
        Return the image, as a PIL object.

        Parameters
        ----------
        image_id : str
        image_variant : str
            Accepted values can be found in the ImageVariants class.
        using_production_images : bool
            Whether to search in production.  

        Returns
        -------
        PIL.Image

        Raises
        ------
        LouvreImageNotFound,
        LouvreInvalidImageVariant,
        LouvreKeyError,
        LouvreQueryError,
        RequestException
        """
        file_url = self.get_image_url(
            image_id=image_id,
            image_variant=image_variant,
            using_production_images=using_production_images)
        _response = requests.get(url=file_url)
        _file_bytes = BytesIO(_response.content)
        return Image.open(_file_bytes)

    def update_image_metadata(self,
                              update_request: UpdateMetadataRequest) -> None:
        """
        Update the metadata for an image.

        Parameters
        ----------
        update_request : UpdateMetadataRequest

        Raises
        ------
        RequestException
        LouvreQueryError
        """
        ImageEnhanceAPI.update_metadata(update_request=update_request)



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union


@dataclass
class ImageFile:
    """Represent image files from ImageAPI."""

    image_variant: str
    size: int
    sasuri: str
    height: float
    width: float


@dataclass
class ImageData:
    """Represent image data from ImageAPI."""

    image_id: str
    etag: str
    metadata: List[dict]
    image_files: List[ImageFile]

    @property
    def image_variants(self) -> List[str]:
        """
        Return a list with the imageVariants present.
        """
        return [image_file.image_variant for image_file in self.image_files]

    def get_variant(self, variant: str) -> Union[ImageFile, None]:
        """
        Given an image variant, return its corresponding ImageFile if it exits in the ImageData 
        instance.

        Parameters
        ----------
        variant : str
            Name of the desired image variant.

        Returns
        -------
        ImageFile, None
        """
        for image_file in self.image_files:
            if variant == image_file.image_variant:
                return image_file
        return None


@dataclass
class UpdateMetadataRequest:
    """Represent a metadata update request to be sent to ImageEnhanceAPI."""

    image_id: str
    additional_metadata: Dict[str, str] = field(default_factory=dict)
    plugin_id: Optional[str] = None
    e_tag: Optional[str] = None
    skip_e_tag_validation: bool = True
    client_name: str = 'python-client'



"""
API wrapper for talking to the ImageAPI to fetch images based on IDs.

Authors: Rafael Sanchez & Sindre Eik de Lange
Date: 26.08.20
"""

from http import HTTPStatus
import requests
from typing import Any, Dict, Union, List

from .config import Config
from .data_models import ImageData, ImageFile
from .enums import ImageVariants, SecretPaths
from .errors import LouvreImageNotFound, LouvreQueryError, LouvreKeyError, \
    LouvreInvalidImageVariant
from .token import Token


class ImageAPI():
    """Wrapper class for communicating with ImageAPI."""
    @staticmethod
    def get_image_data(img_id: str,
                       using_production_images: bool = False) -> ImageData:
        """
        Return image data for a given image ID: parsed additionalMetadata, _etag, imageFiles.

        Parameters
        ----------
        img_id : str
            ID of the image.
        using_production_images : bool
            Whether to search in production.

        Returns
        -------
        ImageData

        Raises
        ------      
        LouvreImageNotFound,
        LouvreKeyError,
        LouvreQueryError,
        RequestException,
        """

        query = \
            f'''query GetImage {{
                    getImage(id: "{img_id}")
                    {{
                        _etag
                        additionalMetadata {{
                            key
                            value          
                        }}
                        imageFiles {{
                            imageVariant
                            sasUri
                            size
                            height
                            width
                        }}                        
                    }}
                }}'''

        response = requests.post(
            url=Config.get_secret(
                SecretPaths.LOUVRE_DOMAIN_PROD if using_production_images else
                SecretPaths.LOUVRE_DOMAIN) + Config.image_api_relative_url,
            json={'query': query},
            headers=Token.get_headers(
                using_production_images=using_production_images))

        if response.status_code == HTTPStatus.OK:
            try:
                return ImageData(
                    image_id=img_id,
                    etag=response.json()['data']['getImage']['_etag'],
                    image_files=ImageAPI._extract_image_files(
                        image_files=response.json()['data']['getImage']
                        ['imageFiles']),
                    metadata=ImageAPI._extract_additional_metadata(
                        metadata=response.json()['data']['getImage']
                        ['additionalMetadata']))
            except KeyError as key_error:
                raise LouvreKeyError(str(key_error))
        elif response.status_code == HTTPStatus.BAD_REQUEST and 'No image with id' in str(
                response.content):
            raise LouvreImageNotFound()
        raise LouvreQueryError()

    @staticmethod
    def get_image_sasuri(img_id: str,
                         image_variant: Union[str, None] = None,
                         using_production_images: bool = False) -> str:
        """
        Given image ID and image variant, return the corresponding sasuri.

        Parameters
        ----------
        img_id : str
            ID of the image.
        image_variant : str, optional
            Image size. Valid choices are thumbnail, standard and original.
        using_production_images : bool
            Whether to search in production.
            
        Returns
        -------
        sasuri : str
            Link to the desired image file with a built-in access token.

        Raises
        ------
        LouvreImageNotFound,
        LouvreKeyError,
        LouvreQueryError,
        LouvreInvalidImageVariant,
        RequestException,
        """

        if image_variant is None:
            image_variant = ImageVariants.DEFAULT
        if image_variant not in [
                ImageVariants.THUMBNAIL, ImageVariants.STANDARD,
                ImageVariants.ORIGINAL
        ]:
            raise LouvreInvalidImageVariant()

        query = \
            f'''query GetImage {{
                    getImage(id: "{img_id}", imageVariants: ["{image_variant}"]) {{
                        imageFiles {{
                            sasUri
                        }}
                    }}
                }}'''

        response = requests.post(
            url=Config.get_secret(
                SecretPaths.LOUVRE_DOMAIN_PROD if using_production_images else
                SecretPaths.LOUVRE_DOMAIN) + Config.image_api_relative_url,
            json={'query': query},
            headers=Token.get_headers(
                using_production_images=using_production_images))

        if response.status_code == 200:
            try:
                return response.json(
                )['data']['getImage']['imageFiles'][0]['sasUri']
            except KeyError as key_error:
                raise LouvreKeyError(str(key_error))
        elif response.status_code == HTTPStatus.BAD_REQUEST and 'No image with id' in str(
                response.content):
            raise LouvreImageNotFound()
        else:
            raise LouvreQueryError()

    @staticmethod
    def _extract_image_files(image_files: list) -> List[ImageFile]:
        result = []
        if isinstance(image_files, list):
            for image_file in image_files:
                if isinstance(image_file, dict) and all([
                        key in image_file.keys() for key in
                    ['imageVariant', 'size', 'sasUri', 'height', 'width']
                ]):
                    result.append(
                        ImageFile(image_variant=str(
                            image_file['imageVariant']),
                                  size=int(image_file['size']),
                                  sasuri=str(image_file['sasUri']),
                                  height=float(image_file['height']),
                                  width=float(image_file['width'])))
        return result

    @staticmethod
    def _extract_additional_metadata(metadata: List[Dict[str, Any]]) -> list:
        """Return a list of dictionaries, each with the original key name and their corresponding values, 
        without deserialising."""
        result: List[Dict[str, Any]] = []
        if isinstance(metadata, list):
            for item in metadata:
                if isinstance(item, dict) and all(
                        _key in item.keys() for _key in ['key', 'value']):
                    result.extend({item['key']: item['value']})
        return result



from typing import Union

from elvia_vault import VaultClient
from .errors import LouvreVaultConnError, LouvreSecretError


class Config():
    """Configuration values and secrets."""

    vault: VaultClient
    secrets: dict = {}

    # ImageAPI - GraphQL API to fetch images
    image_api_relative_url: str = '/image/graphql'

    # ImageEnhanceAPI - API to update image's metadata
    image_enhance_api_relative_url: str = '/imageenhance/Metadata'

    # Token buffer in seconds
    token_buffer_seconds: int = 60

    @classmethod
    def connect(cls, github_token: Union[str, None] = None):
        """        
        Connects to the vault using VAULT_ADDR from os.environ.
        If cluster, it uses kubernetes auth, else the Github token.

        Raises
        ------
        LouvreVaultConnError
        """
        # Clean up, to avoid mixing secrets from different vaults, in the case
        # VAULT_ADDR changes between calls
        cls.secrets = {}
        # Get a fresh VaultClient instance and connect to the vault service
        try:
            cls.vault = VaultClient(github_token=github_token)
        except Exception as exception:
            raise LouvreVaultConnError(str(exception))

    @classmethod
    def get_secret(cls, secret_path: str) -> str:
        """
        Return a secret value given a secret path.
        Fetch from vault only if not previously fetched.

        Parameters
        ----------
        secret_path : str

        Raises
        ------
        LouvreSecretError
        """
        try:
            if not secret_path in cls.secrets.keys():
                cls.secrets[secret_path] = cls.vault.get_value(secret_path)
            return cls.secrets[secret_path]

        except KeyError as exception:
            raise LouvreSecretError(exception)



"""
Module with methods to authenticate against ImageAPI and ImageEnhanceAPI.
"""

from datetime import datetime, timedelta
from typing import Dict
import requests
from .config import Config
from .enums import SecretPaths


class Token():
    """Class for working with jwt tokens."""

    _tokens: Dict[str, str] = {}
    _expiry: Dict[str, datetime] = {}

    @classmethod
    def get_token(cls, using_production_images: bool = False) -> str:
        """Return the current token if still valid, else a new one.
           Assume that the token life has ended when time is not smaller than:
           _expiry minus the buffer"""

        endpoint: str = SecretPaths.TOKEN_ENDPOINT_PROD if using_production_images else SecretPaths.TOKEN_ENDPOINT
        if endpoint in cls._tokens.keys() and endpoint in cls._expiry.keys(
        ) and datetime.now() < cls._expiry[endpoint] - timedelta(
                seconds=Config.token_buffer_seconds):
            return cls._tokens[endpoint]
        return cls.get_new_token(
            using_production_images=using_production_images)

    @classmethod
    def get_new_token(cls, using_production_images: bool = False) -> str:
        """Create and get token for this API to communicate with ImageAPI and ImageEnhanceAPI.
           Assume the expiry time is found by adding expires_in to the time right before triggering the request."""
        tick = datetime.now()
        token_response = requests.post(
            url=Config.get_secret(
                SecretPaths.TOKEN_ENDPOINT_PROD
                if using_production_images else SecretPaths.TOKEN_ENDPOINT),
            headers={},
            data={
                'grant_type':
                'client_credentials',
                'client_id':
                Config.get_secret(
                    SecretPaths.CLIENT_ID_API_PROD
                    if using_production_images else SecretPaths.CLIENT_ID_API),
                'client_secret':
                Config.get_secret(
                    SecretPaths.CLIENT_SECRET_API_PROD if
                    using_production_images else SecretPaths.CLIENT_SECRET_API)
            })

        endpoint: str = SecretPaths.TOKEN_ENDPOINT_PROD if using_production_images else SecretPaths.TOKEN_ENDPOINT
        cls._tokens[endpoint] = token_response.json()['access_token']
        cls._expiry[endpoint] = tick + timedelta(
            seconds=int(token_response.json()['expires_in']))
        return cls._tokens[endpoint]

    @classmethod
    def get_headers(cls, using_production_images: bool = False) -> dict:
        """Return headers to be used in HTTP requests against ImageAPI / ImageEnhanceAPI."""
        return {
            "Authorization":
            "Bearer {}".format(
                cls.get_token(
                    using_production_images=using_production_images)),
            'Content-Type':
            'application/json'
        }



"""
Exception classes.
"""


class LouvreException(Exception):
    pass


class LouvreImageNotFound(LouvreException):
    pass


class LouvreKeyError(LouvreException):
    pass


class LouvreQueryError(LouvreException):
    pass


class LouvreValueError(LouvreException):
    pass


class LouvreVaultConnError(LouvreException):
    pass


class LouvreSecretError(LouvreException):
    pass


class LouvreInvalidImageVariant(LouvreException):
    pass



"""
Module with methods to facilitate communication with ImageEnhanceAPI.
"""

from http import HTTPStatus
import json
import requests

from .data_models import UpdateMetadataRequest
from .enums import SecretPaths
from .errors import LouvreQueryError
from .config import Config
from .token import Token


class ImageEnhanceAPI():
    """
    ImageEnhanceAPI wrapper.
    """
    @staticmethod
    def update_metadata(update_request: UpdateMetadataRequest) -> None:
        """
        Update image metadata. Existing metadata entries with the same keys will get overwritten.

        Parameters
        ----------
        update_request: UpdateMetadataRequest

        Raises
        ------
        RequestException
        LouvreQueryError
        """
        payload = {
            "imageID": update_request.image_id,
            "additionalMetadata": update_request.additional_metadata,
            "skipValidationOfProvidedETag":
            update_request.skip_e_tag_validation,
            "clientName": update_request.client_name
        }

        if update_request.e_tag:
            payload["eTag"] = update_request.e_tag

        if update_request.plugin_id:
            payload["pluginId"] = update_request.plugin_id

        query = json.dumps(payload)

        response = requests.put(
            url=Config.get_secret(SecretPaths.LOUVRE_DOMAIN) +
            Config.image_enhance_api_relative_url,
            headers=Token.get_headers(),
            data=query)

        if not response.status_code == HTTPStatus.OK:
            raise LouvreQueryError(response.content)
